{ *************************************************************************** }
{ FromOpenSSL.pas }
{ for Delphi }
{ }
{ Written by Cyberens Technologies & Services }
{ copyright  2016 - 2017 by Cyberens Technologies & Services }
{ Email : marion.candau@cyberens.fr }
{ Web : http://www.cyberens.fr }
{ }
{ The source code is given as is. The author is not responsible }
{ for any possible damage done due to the use of this code. }
{ The component can be freely used in any application. The complete }
{ source code remains property of the author and may not be distributed, }
{ published, given or sold in any form as such. No parts of the source }
{ code can be included in any other component or application without }
{ written authorization of the author. }
{ *************************************************************************** }

unit FromOpenSSL;

interface

uses
  Classes, IdSSLOpenSSLHeaders;

type
  TFromOpenSSL = class(TObject)
  private
    Fpwd: string;
    FSSLModule: HMODULE;
    function LoadOpenSSLLibrary: Boolean;
    procedure UnLoadOpenSSLLibrary;
    function LoadSSLCryptoLibrary: HMODULE;
  public
    Constructor Create;
    Destructor Destroy; override;
    function FromOpenSSLCert(filePath: string): pRSA;
    function FromOpenSSLPublicKey(filePath: string): pRSA;
    function FromOpenSSLPrivateKey(filePath: string): pRSA;
    function FromOpenSSLEncPrivateKey(filePath, Password: string): pRSA;
  published
    property pwd: string read Fpwd;
  end;

implementation

uses
{$IF defined(IOS) and defined(CPUARM)}IdSSLOpenSSLHeaders_Static, {$IFEND}
  IdSSLOpenSSL, IdGlobal,
{$IFDEF MSWINDOWS}
  Winapi.Windows,
{$ENDIF MSWINDOWS}
{$IFDEF POSIX}
  Posix.Dlfcn, Posix.Fcntl, Posix.SysStat, Posix.SysTime, Posix.SysTypes,
  Posix.Stdio, Posix.Stdlib,
{$ENDIF POSIX}
{$IFDEF PC_MAPPED_EXCEPTIONS}
  System.Internal.Unwinder,
{$ENDIF PC_MAPPED_EXCEPTIONS}
  System.SysConst,
  SysUtils, System.IOUtils;

var

  { variables non dfinies dans IdSSLOpenSSLHeaders }
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  PEM_read_bio_PUBKEY: function(bp: PBIO; x: PPEVP_PKEY; cb: ppem_password_cb;
    u: pointer): PEVP_PKEY; cdecl;
  X509_get_pubkey: function(a: pX509): PEVP_PKEY; cdecl;
{$ELSEIF defined(ANDROID)}
  PEM_read_bio_PUBKEY: function(bp: PBIO; x: PPEVP_PKEY; cb: ppem_password_cb;
    u: pointer): PEVP_PKEY;
  X509_get_pubkey: function(a: pX509): PEVP_PKEY;
{$IFEND}
  { fonctions non dfinies dans IdSSLOpenSSLHeaders }
{$IFDEF IOS}
function PEM_read_bio_PUBKEY(bp: PBIO; x: PPEVP_PKEY; cb: ppem_password_cb;
  u: pointer): PEVP_PKEY; external 'libcrypto.a';
function X509_get_pubkey(a: pX509): PEVP_PKEY; external 'libcrypto.a';
function EVP_PKEY_get1_RSA(pkey: PEVP_PKEY): Prsa_st; external 'libcrypto.a';
{$ENDIF}

Constructor TFromOpenSSL.Create;
var
  s: string;
begin
  if not LoadOpenSSLLibrary then
  begin
    s := IdSSLOpenSSLHeaders.WhichFailedToLoad;
    raise Exception.Create('Impossible de charger OpenSSL : ' + s);
  end;
end;

Destructor TFromOpenSSL.Destroy;
begin
  UnLoadOpenSSLLibrary;
end;

{ pour charger les fonctions crypto d'OpenSSL }
function TFromOpenSSL.LoadSSLCryptoLibrary: HMODULE;
begin
{$IFDEF MSWINDOWS}
  Result := SafeLoadLibrary('libeay32.dll');
{$ELSE}
{$IFDEF ANDROID}
  Result := LoadLibrary('libcrypto.so');
{$ELSE}
{$IFNDEF IOS}
  Result := LoadLibrary(PChar('libcrypto.dylib'));
{$ENDIF}
{$ENDIF}
{$ENDIF}
end;

{ pour charger OpenSSL }
function TFromOpenSSL.LoadOpenSSLLibrary: Boolean;
begin
  if FSSLModule <> 0 then
    Exit(True);
{$IFDEF ANDROID}
  IdOpenSSLSetLibPath(IncludeTrailingPathDelimiter(System.IOUtils.TPath.GetDocumentsPath));
{$ENDIF}
  Result := IdSSLOpenSSL.LoadOpenSSLLibrary;
  if Result then
  begin
{$IFNDEF IOS}
    FSSLModule := LoadSSLCryptoLibrary;
    if FSSLModule = 0 then
      Exit(False);
    PEM_read_bio_PUBKEY := GetProcAddress(FSSLModule,
      PChar('PEM_read_bio_PUBKEY'));
    X509_get_pubkey := GetProcAddress(FSSLModule, PChar('X509_get_pubkey'));
    OpenSSL_add_all_algorithms;
    OpenSSL_add_all_ciphers;
    OpenSSL_add_all_digests;
    ERR_load_crypto_strings;
{$ENDIF}
  end;
end;

procedure TFromOpenSSL.UnLoadOpenSSLLibrary;
begin
  if FSSLModule <> 0 then
    FreeLibrary(FSSLModule);
end;

function TFromOpenSSL.FromOpenSSLCert(filePath: string): pRSA;
var
  Stream: TStream;
  KeyBuffer: PBIO;
  Buffer: TBytes;
  FX509: pX509;
  Key: PEVP_PKEY;
  LEncoding: TEncoding;
  LOffset: Integer;
begin
  Stream := TFileStream.Create(filePath, fmOpenRead or fmShareDenyWrite);
  try
    SetLength(Buffer, Stream.size);
    Stream.ReadBuffer(Buffer[0], Stream.size);
{$IFNDEF MSWINDOWS}
    LOffset := TEncoding.GetBufferEncoding(Buffer, LEncoding);
    Buffer := LEncoding.Convert(LEncoding, TEncoding.UTF8, Buffer, LOffset,
      Length(Buffer) - LOffset);
{$ENDIF}
    KeyBuffer := BIO_new_mem_buf(Buffer, Length(Buffer));
    if KeyBuffer = nil then
      raise Exception.Create('Impossible de charger le buffer X509');
    try
      FX509 := PEM_read_bio_X509(KeyBuffer, nil, nil, nil);
      if not Assigned(FX509) then
        raise Exception.Create('Impossible de charger le certificat X509');
      Key := X509_get_pubkey(FX509);
      if not Assigned(Key) then
        raise Exception.Create('Impossible de charger la cl publique X509');
      try
        Result := EVP_PKEY_get1_RSA(Key);
        if not Assigned(Result) then
          raise Exception.Create('Impossible de charger la cl publique RSA');
      finally
        EVP_PKEY_free(Key);
      end;
    finally
      BIO_free(KeyBuffer);
    end;
  finally
    Stream.free;
  end;
end;

function TFromOpenSSL.FromOpenSSLPublicKey(filePath: string): pRSA;
var
  Stream: TStream;
  KeyBuffer: PBIO;
  pkey: PEVP_PKEY;
  Buffer: TBytes;
  LEncoding: TEncoding;
  LOffset: Integer;
begin
  Stream := TFileStream.Create(filePath, fmOpenRead or fmShareDenyWrite);
  try
    SetLength(Buffer, Stream.size);
    Stream.Read(Buffer[0], Stream.size);
{$IFNDEF MSWINDOWS}
    LOffset := TEncoding.GetBufferEncoding(Buffer, LEncoding);
    Buffer := LEncoding.Convert(LEncoding, TEncoding.UTF8, Buffer, LOffset,
      Length(Buffer) - LOffset);
{$ENDIF}
    KeyBuffer := BIO_new_mem_buf(Buffer, Length(Buffer));
    if KeyBuffer = nil then
      raise Exception.Create('Impossible de charger le buffer');
    try
      pkey := PEM_read_bio_PUBKEY(KeyBuffer, nil, nil, nil);
      if not Assigned(pkey) then
        raise Exception.Create('Impossible de charger la cl publique');
      try
        Result := EVP_PKEY_get1_RSA(pkey);
        if not Assigned(Result) then
          raise Exception.Create('Impossible de charger la cl publique RSA');
      finally
        EVP_PKEY_free(pkey);
      end;
    finally
      BIO_free(KeyBuffer);
    end;
  finally
    Stream.free;
  end;

end;

function TFromOpenSSL.FromOpenSSLPrivateKey(filePath: string): pRSA;
var
  Stream: TStream;
  KeyBuffer: PBIO;
  Buffer: TBytes;
  LEncoding: TEncoding;
  LOffset: Integer;
begin
  Stream := TFileStream.Create(filePath, fmOpenRead or fmShareDenyWrite);
  try
    SetLength(Buffer, Stream.size);
    Stream.ReadBuffer(Buffer[0], Stream.size);
{$IFNDEF MSWINDOWS}
    LOffset := TEncoding.GetBufferEncoding(Buffer, LEncoding);
    Buffer := LEncoding.Convert(LEncoding, TEncoding.UTF8, Buffer, LOffset,
      Length(Buffer) - LOffset);
{$ENDIF}
    KeyBuffer := BIO_new_mem_buf(Buffer, Length(Buffer));
    if KeyBuffer = nil then
      raise Exception.Create('Impossible de charger le buffer');
    try
      Result := PEM_read_bio_RSAPrivateKey(KeyBuffer, nil, nil, nil);
      if not Assigned(Result) then
        raise Exception.Create('Impossible de charger la cl prive RSA');
    finally
      BIO_free(KeyBuffer);
    end;
  finally
    Stream.free;
  end;
end;

{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}

function ReadKeyCallback(buf: PAnsiChar; buffsize: Integer; rwflag: Integer;
  u: pointer): Integer; cdecl;
var
  len: Integer;
  Password: string;
  I: Integer;
  fos: TFromOpenSSL;
begin
  Result := 0;
{on rcupre l'objet courant}
  fos := TFromOpenSSL(u);
  if Assigned(u) then
  begin
{on rcupre le mot de passe prcedemment enregistr}
    Password := fos.pwd;
    if Length(Password) < buffsize then
      len := Length(Password)
    else
      len := buffsize;
{on le copie dans buf}
    for I := 0 to len do
      buf[I] := AnsiChar(Password[I + 1]);
    buf[len] := #0;
    Result := len;
  end;
end;
{$ELSE}

function ReadKeyCallback(buf: PByte; buffsize: Integer; rwflag: Integer;
  u: pointer): Integer; cdecl;
var
  len: Integer;
  Password: string;
  I: Integer;
  fos: TFromOpenSSL;
begin
  Result := 0;
{on rcupre l'objet courant}
  fos := TFromOpenSSL(u);
  if Assigned(u) then
  begin
{on rcupre le mot de passe prcedemment enregistr}
    Password := fos.pwd;
    if Length(Password) < buffsize then
      len := Length(Password)
    else
      len := buffsize;
{on le copie dans buf}
    for I := 0 to len do
      buf[I] := Byte(Password[I]);
    buf[len] := Byte(#0);
    Result := len;
  end;
end;
{$IFEND}

function TFromOpenSSL.FromOpenSSLEncPrivateKey(filePath, Password: string): pRSA;
var
  Stream: TStream;
  KeyBuffer: PBIO;
  Buffer, nBuf: TBytes;
  cb: ppem_password_cb;
  LEncoding: TEncoding;
  LOffset: Integer;
begin
{On enregistre le mot de passe}
  Fpwd := Password;
  Stream := TFileStream.Create(filePath, fmOpenRead or fmShareDenyWrite);
  try
    SetLength(Buffer, Stream.size);
    Stream.ReadBuffer(Buffer[0], Stream.size);
{$IFNDEF MSWINDOWS}
    LOffset := TEncoding.GetBufferEncoding(Buffer, LEncoding);
    Buffer := LEncoding.Convert(LEncoding, TEncoding.UTF8, Buffer, LOffset,
      Length(Buffer) - LOffset);
{$ENDIF}
    KeyBuffer := BIO_new_mem_buf(Buffer, Length(Buffer));
    if KeyBuffer = nil then
      raise Exception.Create('Impossible de charger le buffer');
    try
      cb := @ReadKeyCallback;
      Result := PEM_read_bio_RSAPrivateKey(KeyBuffer, nil, cb, nil);
      if not Assigned(Result) then
        raise Exception.Create('Impossible de charger la cl prive RSA');
    finally
      BIO_free(KeyBuffer);
    end;
  finally
    Stream.free;
{on efface le mot de passe}
    Fpwd := '';
  end;
end;

end.
