{ *************************************************************************** }
{ FromOpenSSL.pas }
{ for Delphi }
{ }
{ Written by Cyberens Technologies & Services }
{ copyright  2016 - 2017 by Cyberens Technologies & Services }
{ Email : marion.candau@cyberens.fr }
{ Web : http://www.cyberens.fr }
{ }
{ The source code is given as is. The author is not responsible }
{ for any possible damage done due to the use of this code. }
{ The component can be freely used in any application. The complete }
{ source code remains property of the author and may not be distributed, }
{ published, given or sold in any form as such. No parts of the source }
{ code can be included in any other component or application without }
{ written authorization of the author. }
{ *************************************************************************** }

unit FromOpenSSL;

interface

uses
  Classes, IdSSLOpenSSLHeaders;

{$POINTERMATH ON}

type

{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  ReadKeyChar = AnsiChar;
{$ELSE}
  ReadKeyChar = Byte;
{$ENDIF}
  PReadKeyChar = ^ReadKeyChar;

  TFromOpenSSL = class(TObject)
  private
    FSSLModule: HMODULE;
    function LoadOpenSSLLibrary: Boolean;
    procedure UnLoadOpenSSLLibrary;
    function LoadSSLCryptoLibrary: HMODULE;
  public
    Constructor Create;
    Destructor Destroy; override;
    function FromOpenSSLCert(filePath: string): pRSA;
    function FromOpenSSLPublicKey(filePath: string): pRSA;
    function FromOpenSSLPrivateKey(filePath: string; pwd: String = ''): pRSA;
    function LoadPEMFile(filePath: string): PBio;
  end;

implementation

uses
{$IF defined(IOS) and defined(CPUARM)}IdSSLOpenSSLHeaders_Static, {$IFEND}
  IdSSLOpenSSL, IdGlobal,
{$IFDEF MSWINDOWS}
  Winapi.Windows,
{$ENDIF MSWINDOWS}
{$IFDEF POSIX}
  Posix.Dlfcn, Posix.Fcntl, Posix.SysStat, Posix.SysTime, Posix.SysTypes,
  Posix.Stdio, Posix.Stdlib,
{$ENDIF POSIX}
{$IFDEF PC_MAPPED_EXCEPTIONS}
  System.Internal.Unwinder,
{$ENDIF PC_MAPPED_EXCEPTIONS}
  System.SysConst,
  SysUtils, System.IOUtils, ImportOpenSSLFile;

{ variables non dfinies dans IdSSLOpenSSLHeaders }
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}

var
  PEM_read_bio_PUBKEY: function(bp: PBio; x: PPEVP_PKEY; cb: ppem_password_cb;
    u: pointer): PEVP_PKEY; cdecl;
  X509_get_pubkey: function(a: pX509): PEVP_PKEY; cdecl;
  EVP_PKEY_get1_RSA: function(pkey: PEVP_PKEY): pRSA; cdecl;
{$ELSEIF defined(ANDROID)}

var
  PEM_read_bio_PUBKEY: function(bp: PBio; x: PPEVP_PKEY; cb: ppem_password_cb;
    u: pointer): PEVP_PKEY;
  X509_get_pubkey: function(a: pX509): PEVP_PKEY;
  EVP_PKEY_get1_RSA: function(pkey: PEVP_PKEY): pRSA;
{$IFEND}
  { fonctions non dfinies dans IdSSLOpenSSLHeaders }
{$IFDEF IOS}
function PEM_read_bio_PUBKEY(bp: PBio; x: PPEVP_PKEY; cb: ppem_password_cb;
  u: pointer): PEVP_PKEY; external 'libcrypto.a';
function X509_get_pubkey(a: pX509): PEVP_PKEY; external 'libcrypto.a';
function EVP_PKEY_get1_RSA(pkey: PEVP_PKEY): pRSA; external 'libcrypto.a';
{$ENDIF}

Constructor TFromOpenSSL.Create;
var
  s: string;
begin
{$IFNDEF IOS}
  if not LoadOpenSSLLibrary then
  begin
    s := IdSSLOpenSSLHeaders.WhichFailedToLoad;
    raise Exception.Create('Impossible de charger OpenSSL : ' + s);
  end;
{$ENDIF}
end;

Destructor TFromOpenSSL.Destroy;
begin
  UnLoadOpenSSLLibrary;
  inherited;
end;

{ pour charger les fonctions crypto d'OpenSSL }
function TFromOpenSSL.LoadSSLCryptoLibrary: HMODULE;
begin
{$IFDEF MSWINDOWS}
  Result := SafeLoadLibrary('libeay32.dll');
{$ELSE}
{$IFDEF ANDROID}
  Result := LoadLibrary('libcrypto.so');
{$ELSE}
{$IFNDEF IOS}
  Result := LoadLibrary('libcrypto.dylib');
{$ENDIF}
{$ENDIF}
{$ENDIF}
end;

{ pour charger OpenSSL }
function TFromOpenSSL.LoadOpenSSLLibrary: Boolean;
begin
  if FSSLModule <> 0 then
    Exit(True);
{$IFDEF ANDROID}
  IdOpenSSLSetLibPath(System.IOUtils.TPath.GetDocumentsPath);
{$ENDIF}
  Result := IdSSLOpenSSL.LoadOpenSSLLibrary;
  if Result then
  begin
{$IFNDEF IOS}
    FSSLModule := LoadSSLCryptoLibrary;
    if FSSLModule = 0 then
      Exit(False);
    PEM_read_bio_PUBKEY := GetProcAddress(FSSLModule,
      PChar('PEM_read_bio_PUBKEY'));
    X509_get_pubkey := GetProcAddress(FSSLModule, PChar('X509_get_pubkey'));
    EVP_PKEY_get1_RSA := GetProcAddress(FSSLModule, PChar('EVP_PKEY_get1_RSA'));
    OpenSSL_add_all_algorithms;
    OpenSSL_add_all_ciphers;
    OpenSSL_add_all_digests;
    ERR_load_crypto_strings;
{$ENDIF}
  end;
end;

procedure TFromOpenSSL.UnLoadOpenSSLLibrary;
begin
  if FSSLModule <> 0 then
    FreeLibrary(FSSLModule);
end;

function TFromOpenSSL.LoadPEMFile(filePath: string): PBio;
var
{$IFNDEF MSWINDOWS}
  LEncoding: TEncoding;
  LOffset: Integer;
{$ENDIF}
  Buffer: TBytes;
  Stream: TStream;
begin
  Stream := TFileStream.Create(filePath, fmOpenRead or fmShareDenyWrite);
  try
    SetLength(Buffer, Stream.size);
    Stream.ReadBuffer(Buffer[0], Stream.size);
{$IFNDEF MSWINDOWS}
    LEncoding := nil;
    LOffset := TEncoding.GetBufferEncoding(Buffer, LEncoding);
    Buffer := LEncoding.Convert(LEncoding, TEncoding.UTF8, Buffer, LOffset,
      Length(Buffer) - LOffset);
{$ENDIF}
    Result := BIO_new_mem_buf(Buffer, Length(Buffer));
  finally
    Stream.free;
  end;
end;

function TFromOpenSSL.FromOpenSSLCert(filePath: string): pRSA;
var
  KeyBuffer: PBio;
  FX509: pX509;
  Key: PEVP_PKEY;

begin
  KeyBuffer := LoadPEMFile(filePath);
  if KeyBuffer = nil then
    raise Exception.Create('Impossible de charger le buffer X509');
  try
    FX509 := PEM_read_bio_X509(KeyBuffer, nil, nil, nil);
    if not Assigned(FX509) then
      raise Exception.Create('Impossible de charger le certificat X509');
    Key := X509_get_pubkey(FX509);
    if not Assigned(Key) then
      raise Exception.Create('Impossible de charger la cl publique X509');
    try
      Result := EVP_PKEY_get1_RSA(Key);
      if not Assigned(Result) then
        raise Exception.Create('Impossible de charger la cl publique RSA');
    finally
      EVP_PKEY_free(Key);
    end;
  finally
    BIO_free(KeyBuffer);
  end;

end;

function TFromOpenSSL.FromOpenSSLPublicKey(filePath: string): pRSA;
var
  KeyBuffer: PBio;
  pkey: PEVP_PKEY;
begin
  KeyBuffer := LoadPEMFile(filePath);
  if KeyBuffer = nil then
    raise Exception.Create('Impossible de charger le buffer');
  try
    pkey := PEM_read_bio_PUBKEY(KeyBuffer, nil, nil, nil);
    if not Assigned(pkey) then
      raise Exception.Create('Impossible de charger la cl publique');
    try
      Result := EVP_PKEY_get1_RSA(pkey);
      if not Assigned(Result) then
        raise Exception.Create('Impossible de charger la cl publique RSA');
    finally
      EVP_PKEY_free(pkey);
    end;
  finally
    BIO_free(KeyBuffer);
  end;
end;

function TFromOpenSSL.FromOpenSSLPrivateKey(filePath: string;
  pwd: String): pRSA;
var
  KeyBuffer: PBio;
  p: PReadKeyChar;
  I: Integer;
begin
  KeyBuffer := LoadPEMFile(filePath);
  if KeyBuffer = nil then
    raise Exception.Create('Impossible de charger le buffer');
  try
    if pwd <> '' then
    begin
      p := GetMemory((pwd.Length + 1) * SizeOf(Char));
      for I := 0 to pwd.Length - 1 do
        p[I] := ReadKeyChar(pwd.Chars[I]);
      p[pwd.Length] := ReadKeyChar(#0);
    end
    else
      p := nil;
    try
      Result := PEM_read_bio_RSAPrivateKey(KeyBuffer, nil, nil, p);
      if not Assigned(Result) then
        raise Exception.Create('Impossible de charger la cl prive RSA');
    finally
      FillChar(p, SizeOf(p), 0);
      FreeMem(p);
    end;
  finally
    BIO_free(KeyBuffer);
  end;

end;

end.
